package fetcher

import (
	"context"
	"log/slog"
	"time"

	"github.com/google/uuid"
	"gopkg.in/guregu/null.v4"

	"borodyadka.dev/borodyadka/minireader/internal/minireader/dto"
)

type Fetcher interface {
	Fetch(ctx context.Context, source string) (dto.Feed, []dto.Item, error)
}

type feedRepository interface {
	ListOutdatedFeeds(ctx context.Context) ([]dto.Feed, error)
	UpdateFeed(ctx context.Context, input *dto.UpdateFeed) error
	SaveFeedError(ctx context.Context, input *dto.UpdateFeedError) error
}

type itemRepository interface {
	UpsertItems(ctx context.Context, feed uuid.UUID, items []dto.Item) (inserted int, err error)
}

type Config struct {
	CheckInterval time.Duration

	Fetchers map[dto.Provider]Fetcher
}

type FetchScheduler struct {
	checkInterval time.Duration
	logger        *slog.Logger

	feeds feedRepository
	items itemRepository

	fetchers map[dto.Provider]Fetcher
}

func New(opts ...Option) *FetchScheduler {
	fetcher := &FetchScheduler{
		fetchers: make(map[dto.Provider]Fetcher),
	}

	for _, opt := range opts {
		opt(fetcher)
	}

	return fetcher
}

func (s *FetchScheduler) Start(ctx context.Context) error {
	t := time.NewTicker(s.checkInterval)
	defer t.Stop()

	for {
		select {
		case <-ctx.Done():
			return nil
		case <-t.C:
			if err := s.refreshFeeds(ctx); err != nil {
				s.logger.Error("failed to refresh feeds", "error", err)
			}
		}
	}
}

func (s *FetchScheduler) refreshFeeds(ctx context.Context) error {
	logger := s.logger.With("method", "refreshFeeds")

	// TODO: make it parallel with N (configurable) workers
	feeds, err := s.feeds.ListOutdatedFeeds(ctx)
	if err != nil {
		return err
	}

	for _, feed := range feeds {
		err := s.refreshFeed(ctx, feed)
		if err != nil {
			logger.Error("failed to refresh feed", "feed", feed.ID, "error", err)
		}
	}

	return nil
}

func (s *FetchScheduler) refreshFeed(ctx context.Context, feed dto.Feed) error {
	logger := s.logger.With("method", "refreshFeeds")

	fetcher := s.fetchers[feed.Provider]
	if fetcher == nil {
		logger.Error("fetcher not exists", "feed", feed.ID, "provider", feed.Provider, "source", feed.Source)
		return nil
	}
	now := time.Now()

	info, items, err := fetcher.Fetch(ctx, feed.Source)
	if err != nil {
		return s.feeds.SaveFeedError(ctx, &dto.UpdateFeedError{
			ID:            feed.ID,
			Error:         err.Error(),
			LastFetchedAt: now,
		})
	}

	inserted, err := s.items.UpsertItems(ctx, feed.ID, items)
	if err != nil {
		return err
	}

	// find out how often we should refresh this feed:
	// * if ratio of new items is greater than 90% then shorten interval in half
	// * if ratio of new items is less than 10% then double interval
	multiplier := 1.0
	ratio := float64(inserted) / float64(len(items))
	if ratio >= 0.9 {
		multiplier = 0.5
	} else if ratio <= 0.1 {
		multiplier = 2
	}

	logger.Debug("items upserted",
		"feed", feed.ID,
		"provider", feed.Provider,
		"source", feed.Source,
		"ratio", ratio,
		"items", inserted,
		"total", len(items))

	if err := s.feeds.UpdateFeed(ctx, &dto.UpdateFeed{
		ID:                      feed.ID,
		Title:                   null.StringFrom(info.Title),
		FetchIntervalMultiplier: multiplier,
		LastFetchedAt:           now,
	}); err != nil {
		return err
	}

	return nil
}

func (s *FetchScheduler) ForceRefresh(ctx context.Context, feed dto.Feed) error {
	return s.refreshFeed(ctx, feed)
}
