package repositories

import (
	"context"
	"errors"

	"github.com/jackc/pgx/v5"
	"github.com/jackc/pgx/v5/pgconn"
	"github.com/jackc/pgx/v5/pgxpool"
)

type ctxkey string

const (
	ConnKey ctxkey = "pg_conn"
)

var (
	ErrCannotStartTransaction = errors.New("cannot start transaction (method not supported)")
)

type txbeginner interface {
	Begin(ctx context.Context) (pgx.Tx, error)
}

type Conn interface {
	Query(ctx context.Context, sql string, args ...any) (pgx.Rows, error)
	QueryRow(ctx context.Context, sql string, args ...any) pgx.Row
	Exec(ctx context.Context, sql string, arguments ...any) (pgconn.CommandTag, error)
	SendBatch(context.Context, *pgx.Batch) pgx.BatchResults
}

func NewDatabase(ctx context.Context, dsn string) (*pgxpool.Pool, error) {
	cfg, err := pgxpool.ParseConfig(dsn)
	if err != nil {
		return nil, err
	}
	// TODO: logger

	return pgxpool.NewWithConfig(ctx, cfg)
}

func ConnFromContext(ctx context.Context) Conn {
	return ctx.Value(ConnKey).(Conn)
}

func WithConn(ctx context.Context, conn Conn) context.Context {
	return context.WithValue(ctx, ConnKey, conn)
}

type TransactionalFn func(ctx context.Context) error

func Transactional(ctx context.Context, fn TransactionalFn) error {
	conn := ConnFromContext(ctx)

	// already in transaction
	if _, ok := conn.(*pgxpool.Tx); ok {
		return fn(ctx)
	}

	b, ok := conn.(txbeginner)
	if !ok {
		return ErrCannotStartTransaction
	}

	tx, err := b.Begin(ctx)
	if err != nil {
		return err
	}
	defer tx.Rollback(ctx)

	if err := fn(WithConn(ctx, tx)); err != nil {
		return err
	}

	return tx.Commit(ctx)
}
