package repositories

import (
	"context"

	"github.com/google/uuid"
	"github.com/jackc/pgx/v5/pgtype"

	"borodyadka.dev/borodyadka/minireader/internal/minireader/dto"
	"borodyadka.dev/borodyadka/minireader/internal/minireader/repositories/sql"
	"borodyadka.dev/borodyadka/minireader/internal/pkg/utils"
)

type FeedRepository struct {
}

func NewFeedRepository() *FeedRepository {
	return &FeedRepository{}
}

func (r *FeedRepository) FindOrCreate(ctx context.Context, input *dto.CreateFeed) (dto.Feed, bool, error) {
	result, err := sql.New(ConnFromContext(ctx)).FindOrCreateFeed(ctx, sql.FindOrCreateFeedParams{
		Title:    input.Title,
		Provider: input.Provider,
		Source:   input.Source,
	})
	if err != nil {
		return dto.Feed{}, false, err
	}

	return dto.Feed{
		ID:       result.ID,
		Title:    result.Title,
		Provider: input.Provider,
		Source:   input.Source,
	}, result.Created, nil
}

func (r *FeedRepository) ListOutdatedFeeds(ctx context.Context) ([]dto.Feed, error) {
	items, err := sql.New(ConnFromContext(ctx)).ListOutdatedFeeds(ctx)
	if err != nil {
		return nil, err
	}

	return utils.MapSlice(items, func(t sql.ListOutdatedFeedsRow) dto.Feed {
		return dto.Feed{
			ID:       t.ID,
			Provider: t.Provider,
			Source:   t.Source,
		}
	}), nil
}

func (r *FeedRepository) ListAllUserFeeds(ctx context.Context, user *dto.User) ([]dto.Feed, int, error) {
	items, err := sql.New(ConnFromContext(ctx)).ListAllUserFeeds(ctx, user.ID)
	if err != nil {
		return nil, 0, err
	}

	return utils.MapSlice(items, func(t sql.ListAllUserFeedsRow) dto.Feed {
		return dto.Feed{
			ID:          t.ID,
			Title:       t.Title,
			Provider:    t.Provider,
			Source:      t.Source,
			UnreadCount: int(t.Unread),
			State:       dto.FeedState{},
		}
	}), len(items), nil
}

func (r *FeedRepository) UpdateFeed(ctx context.Context, input *dto.UpdateFeed) error {
	return sql.New(ConnFromContext(ctx)).UpdateFeed(ctx, sql.UpdateFeedParams{
		ID:                 input.ID,
		Title:              input.Title.String,
		IntervalMultiplier: input.FetchIntervalMultiplier,
		LastFetchedAt:      pgtype.Timestamptz{Time: input.LastFetchedAt, Valid: true},
	})
}

func (r *FeedRepository) SaveFeedError(ctx context.Context, input *dto.UpdateFeedError) error {
	return sql.New(ConnFromContext(ctx)).SaveFeedError(ctx, sql.SaveFeedErrorParams{
		ID:            input.ID,
		LastError:     input.Error,
		LastFetchedAt: pgtype.Timestamptz{Time: input.LastFetchedAt, Valid: true},
	})
}

func (r *FeedRepository) Delete(ctx context.Context, id uuid.UUID) (bool, error) {
	tag, err := sql.New(ConnFromContext(ctx)).DeleteFeed(ctx, id)
	return tag.RowsAffected() > 0, err
}
