package http

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"

	jwtware "github.com/gofiber/contrib/jwt"
	"github.com/gofiber/fiber/v2"
	"github.com/gofiber/fiber/v2/log"
	"github.com/golang-jwt/jwt/v5"

	"borodyadka.dev/borodyadka/minireader/internal/minireader/config"
	"borodyadka.dev/borodyadka/minireader/internal/minireader/dto"
)

type ctxKey string

const userCtxKey ctxKey = "__user__"

func GetUserFromContext(ctx context.Context) *dto.User {
	user, ok := ctx.Value(userCtxKey).(*dto.User)
	if !ok {
		return nil
	}
	return user
}

func fetchUserID(ctx context.Context, url, token string) (string, error) {
	cl := http.DefaultClient
	req, err := http.NewRequestWithContext(ctx, "GET", url, nil)
	if err != nil {
		return "", err
	}
	req.Header.Add(fiber.HeaderAuthorization, fmt.Sprintf("Bearer %s", token))

	resp, err := cl.Do(req)
	if err != nil {
		return "", err
	}
	defer resp.Body.Close()

	var info struct {
		Sub string `json:"sub"`
	}

	if err := json.NewDecoder(resp.Body).Decode(&info); err != nil {
		return "", err
	}

	return info.Sub, nil
}

func authMiddleware(security config.Security) fiber.Handler {
	return jwtware.New(jwtware.Config{
		SuccessHandler: func(ctx *fiber.Ctx) error {
			claims := ctx.Locals("claims")

			tok, ok := claims.(*jwt.Token)
			if !ok {
				return errors.New("unauthorized") // TODO: normal error
			}

			var sub string
			var err error
			if security.UserInfoURL != "" {
				sub, err = fetchUserID(ctx.Context(), security.UserInfoURL, tok.Raw)
			} else {
				sub, err = tok.Claims.GetSubject()
			}
			if err != nil {
				return err
			}

			if sub == "" {
				log.Error("unable to fetch user ID nor from token nor from auth provider")
				return errors.New("unauthorized") // TODO: normal error
			}

			ctx.Locals(userCtxKey, &dto.User{
				ID: sub,
			})
			return ctx.Next()
		},
		ContextKey: "claims",
		JWKSetURLs: []string{security.JWKSURL},
	})
}
