package http

import (
	"context"
	"log/slog"

	"github.com/gofiber/fiber/v2"
	slogfiber "github.com/samber/slog-fiber"

	"borodyadka.dev/borodyadka/minireader/internal/minireader/config"
	"borodyadka.dev/borodyadka/minireader/internal/minireader/dto"
)

type reader interface {
	EnsureUser(ctx context.Context, user *dto.User) error

	ListAllUserFeeds(ctx context.Context, user *dto.User) ([]dto.Feed, int, error)
	CreateSubscription(ctx context.Context, user *dto.User, input *dto.CreateFeed) (dto.Feed, error)
}

type Server struct {
	server     *fiber.App
	config     config.HTTP
	sec        config.Security
	reader     reader
	logger     *slog.Logger
	injections [][2]any
}

func New(opts ...Option) *Server {
	server := &Server{
		server: fiber.New(fiber.Config{
			ReadTimeout:                  0,   // TODO
			WriteTimeout:                 0,   // TODO
			IdleTimeout:                  0,   // TODO
			ReadBufferSize:               0,   // TODO
			WriteBufferSize:              0,   // TODO
			ErrorHandler:                 nil, // TODO
			DisableStartupMessage:        true,
			DisablePreParseMultipartForm: true,
			EnableSplittingOnParsers:     false,
		}),
	}

	for _, opt := range opts {
		opt(server)
	}

	server.init()

	return server
}

func (s *Server) Inject(key, value any) {
	s.injections = append(s.injections, [2]any{key, value})
}

func (s *Server) init() {
	s.server.Use(func(c *fiber.Ctx) error {
		for _, v := range s.injections {
			c.Locals(v[0], v[1])
		}
		return c.Next()
	})

	if s.config.AccessLog {
		s.server.Use(slogfiber.New(s.logger))
	}

	if s.sec.JWKSURL != "" {
		s.server.Use(authMiddleware(s.sec))
	}

	api := s.server.Group(s.config.APIPrefix)
	v1 := api.Group("/v1")

	user := v1.Group("/me")
	user.Get("/", s.UserInfo) // TODO: init user if not exists

	feeds := v1.Group("/feeds")
	feeds.Get("/", s.ListAllUserFeeds)
	feeds.Post("/", s.CreateNewFeed)
	feeds.Post("/_import/opml", s.ImportOPML)
	feeds.Get("/:feed", s.GetFeedInfo) // there is some special :feed keys, such as "_all"
	feeds.Patch("/:feed", s.UpdateFeedInfo)
	feeds.Delete("/:feed", s.DeleteFeed)

	items := feeds.Group("/:feed/items")
	items.Get("/", s.ListFeedItems)
	items.Get("/:item", s.GetItemInfo)
	items.Put("/:item/readed", s.MarkItemReaded)
	items.Delete("/:item/readed", s.MarkItemUnreaded)
}

func (s *Server) Start() error {
	s.logger.Info("listen", "addr", s.config.Listen)
	return s.server.Listen(s.config.Listen)
}

func (s *Server) Shutdown() error {
	s.logger.Info("shutdown")
	return s.server.Shutdown()
}
